/**********************************************************
 * view.c - view transformation module
 *
 * 1994 by Patrick J. Angeles
 *
 *
 **********************************************************/

#include <stdio.h>
#include <math.h>

#include "matrix.h"
#include "view.h"
#include "main.h"

#define RAD(x) (M_PI/180.0* (x))

/********** DEBUGing flags...
#define DEBUG_CLIP
#define DEBUG_VIEW
#define DEBUG_PIPE
*/

/********** THE FOLLOWING FOR INTERNAL USE ONLY ***********/
extern short    point_clip();
extern short    vector_clip();

void            print_vrp(Transformation *View)
{
   printf("vrp:\n");
   print_point(&(View->view_ref));
   printf("right vector:\n");
   print_vector(View->right_v);
   printf("up vector:\n");
   print_vector(View->up_v);
   printf("normal vector:\n");
   print_vector(View->normal_v);
}

void            update_view_info(Transformation *View,
                                 double theta, double elev, double forw,
                                 double *Location)
{
   point         *view_ref  = &(View->view_ref);
   double        *normal_v  = View->normal_v;
   double        *right_v   = View->right_v;
   double        *up_v      = View->up_v;

   double         Angle;
   double         sin_Angle;
   double         cos_Angle;

   double         inv_Meg;

#ifdef SUN
   Angle = theta / 180.0;
   sincospi(Angle, &sin_Angle, &cos_Angle);
#else
   Angle = M_PI * (theta / 180.0);
   sin_Angle = sin(Angle);
   cos_Angle = cos(Angle);
#endif

   up_v[0] = sin_Angle;
   up_v[1] = tan(M_PI * (75.0 / 180.0));
   up_v[2] = cos_Angle;
   up_v[3] = 1.0;

   inv_Meg = 1.0 / sqrt((1.0 + up_v[1] * up_v[1]));
   up_v[0] *= inv_Meg;
   up_v[1] *= inv_Meg;
   up_v[2] *= inv_Meg;

   normal_v[0] = sin_Angle;
   normal_v[1] = -tan(M_PI * (15.0 / 180.0));
   normal_v[2] = cos_Angle;
   normal_v[3] = 1.0;

   inv_Meg = 1.0 / sqrt((1.0 + normal_v[1] * normal_v[1]));
   normal_v[0] *= inv_Meg;
   normal_v[1] *= inv_Meg;
   normal_v[2] *= inv_Meg;

   normalized_cross_product(right_v, normal_v, up_v);

   if (Location == NULL)
   {
      view_ref->x += -right_v[1] * elev + -right_v[2] * forw;
      view_ref->y += up_v[1] * elev + up_v[2] * forw;
      view_ref->z += normal_v[1] * elev + normal_v[2] * forw;
   }
   else
   {
      view_ref->x = Location[0];
      view_ref->y = Location[1];
      view_ref->z = Location[2];
   }
}

void            compute_transformation(Transformation *View)
{
   register point         *view_ref  = &(View->view_ref);
   register double        *normal_v  = View->normal_v;
   register double        *right_v   = View->right_v;
   register double        *up_v      = View->up_v;

   register matrix        *T_matrix     = &(View->T_matrix);
   register double        *Matrix_row_0 = View->pp_matrix[0];
   register double        *Matrix_row_1 = View->pp_matrix[1];
   register double        *Matrix_row_2 = View->pp_matrix[2];
   register double        *Matrix_row_3 = View->pp_matrix[3];

            double         Vector[3];

#ifdef DEBUG_VIEW
   printf("view orientation matrix:\n");
#endif

   (*T_matrix)[0][0] = Matrix_row_0[0] * -right_v[0]
                       + Matrix_row_0[1] * -right_v[1]
                       + Matrix_row_0[2] * -right_v[2];

   (*T_matrix)[1][0] = Matrix_row_1[0] * -right_v[0]
                       + Matrix_row_1[1] * -right_v[1]
                       + Matrix_row_1[2] * -right_v[2];

   (*T_matrix)[2][0] = Matrix_row_2[0] * -right_v[0]
                       + Matrix_row_2[1] * -right_v[1]
                       + Matrix_row_2[2] * -right_v[2];

   (*T_matrix)[3][0] = Matrix_row_3[0] * -right_v[0]
                       + Matrix_row_3[1] * -right_v[1]
                       + Matrix_row_3[2] * -right_v[2];
#ifdef DEBUG_VIEW
   printf("%lf %lf %lf %lf\n", -right_v[0], -right_v[1], -right_v[2], 0.0);
#endif   

   (*T_matrix)[0][1] = Matrix_row_0[0] * up_v[0]
                       + Matrix_row_0[1] * up_v[1]
                       + Matrix_row_0[2] * up_v[2];

   (*T_matrix)[1][1] = Matrix_row_1[0] * up_v[0]
                       + Matrix_row_1[1] * up_v[1]
                       + Matrix_row_1[2] * up_v[2];

   (*T_matrix)[2][1] = Matrix_row_2[0] * up_v[0]
                       + Matrix_row_2[1] * up_v[1]
                       + Matrix_row_2[2] * up_v[2];

   (*T_matrix)[3][1] = Matrix_row_3[0] * up_v[0]
                       + Matrix_row_3[1] * up_v[1]
                       + Matrix_row_3[2] * up_v[2];
#ifdef DEBUG_VIEW
   printf("%lf %lf %lf %lf\n", up_v[0], up_v[1], up_v[2], 0.0);
#endif   

   (*T_matrix)[0][2] = Matrix_row_0[0] * normal_v[0]
                       + Matrix_row_0[1] * normal_v[1]
                       + Matrix_row_0[2] * normal_v[2];

   (*T_matrix)[1][2] = Matrix_row_1[0] * normal_v[0]
                       + Matrix_row_1[1] * normal_v[1]
                       + Matrix_row_1[2] * normal_v[2];

   (*T_matrix)[2][2] = Matrix_row_2[0] * normal_v[0]
                       + Matrix_row_2[1] * normal_v[1]
                       + Matrix_row_2[2] * normal_v[2];

   (*T_matrix)[3][2] = Matrix_row_3[0] * normal_v[0]
                       + Matrix_row_3[1] * normal_v[1]
                       + Matrix_row_3[2] * normal_v[2];
#ifdef DEBUG_VIEW
   printf("%lf %lf %lf %lf\n", normal_v[0], normal_v[1], normal_v[2], 0.0);
#endif   

   Vector[0] = -right_v[0] * -view_ref->x + up_v[0] * -view_ref->y
               + normal_v[0] * -view_ref->z;
   Vector[1] = -right_v[1] * -view_ref->x + up_v[1] * -view_ref->y
               + normal_v[1] * -view_ref->z;
   Vector[2] = -right_v[2] * -view_ref->x + up_v[2] * -view_ref->y
               + normal_v[2] * -view_ref->z;

   (*T_matrix)[0][3] = Matrix_row_0[0] * Vector[0]
                       + Matrix_row_0[1] * Vector[1]
                       + Matrix_row_0[2] * Vector[2]
                       + Matrix_row_0[3];

   (*T_matrix)[1][3] = Matrix_row_1[0] * Vector[0]
                       + Matrix_row_1[1] * Vector[1]
                       + Matrix_row_1[2] * Vector[2]
                       + Matrix_row_1[3];

   (*T_matrix)[2][3] = Matrix_row_2[0] * Vector[0]
                       + Matrix_row_2[1] * Vector[1]
                       + Matrix_row_2[2] * Vector[2]
                       + Matrix_row_2[3];

   (*T_matrix)[3][3] = Matrix_row_3[0] * Vector[0]
                       + Matrix_row_3[1] * Vector[1]
                       + Matrix_row_3[2] * Vector[2]
                       + Matrix_row_3[3];
#ifdef DEBUG_VIEW
   printf("%lf %lf %lf %lf\n", Vector[0], Vector[1], Vector[2], 1.0);
#endif   
}

/*****************************************************
 * void eval_projection( prp,F,B,frustrum)
 *   D - depth of projection: prp = <0.0 0.0 D>.
 *   F - front clipping plane.
 *   B - back clipping plane.
 *   frustrum - in degrees, scope of vision
 * Evaluates the perspective projection matrix (pp_matrix).
 * Simplified version of Foley, VanDam projection
 *  matrix. Check it for details!
 ****************************************************/
void            eval_projection(Transformation *View,
                                double D, double F, double B, double frustrum)
{
   matrix        *pp_matrix = &(View->pp_matrix);
   double          umin, umax, vmin, vmax;

   /*******************************************
    * Compute umax, etc. in terms of frustrum.
    *******************************************/
#ifdef SUN
   umax = vmax = D * tanpi(frustrum / 360.0);
#else
   umax = vmax = D * tan(M_PI * (frustrum / 360.0));
#endif

   umin = vmin = -umax;

   (*pp_matrix)[0][0] = (2.0 * -D) / ((umax - umin) * (B - D));
   (*pp_matrix)[0][1] = 0.0;
   (*pp_matrix)[0][2] = 0.0;
   (*pp_matrix)[0][3] = 0.0;

   (*pp_matrix)[1][0] = 0.0;
   (*pp_matrix)[1][1] = (2.0 * -D) / ((vmax - vmin) * (B - D));
   (*pp_matrix)[1][2] = 0.0;
   (*pp_matrix)[1][3] = 0.0;

   (*pp_matrix)[2][0] = 0.0;
   (*pp_matrix)[2][1] = 0.0;
   (*pp_matrix)[2][2] = -1.0 / (B - D);
   (*pp_matrix)[2][3] = D / (B - D);

   (*pp_matrix)[3][0] = 0.0;
   (*pp_matrix)[3][1] = 0.0;
   (*pp_matrix)[3][2] = 0.0;
   (*pp_matrix)[3][3] = 1.0;

   View->Zmin = -(F - D) / (B - D);
   View->Zdepth = -D;

#ifdef DEBUG_VIEW
   printf("perspective projection matrix:\n");
   print_matrix(*pp_matrix);
#endif
}

/************************************************************
 * void eval_viewport( vp_minx,vp_maxx,vp_miny,vp_maxy,
 *                     vp_minz,vp_maxz)
 *   - min and max parameters in screen coordinates
 * Maps the viewing window onto the screen window (vp_matrix).
 ***********************************************************/
void            eval_viewport(Transformation *View,
                              double vp_minx, double vp_maxx,
                              double vp_miny, double vp_maxy,
                              double vp_minz, double vp_maxz)
{
   matrix        *vp_matrix = &(View->vp_matrix);
   double         Zmin      = View->Zmin;
   double         Factor;

   /****************************************
    * First, convert canonical view volume
    *  from perspective to parallel.
    ****************************************/
   Factor = (0.0 - Zmin) / (1.0 + Zmin);
   (*vp_matrix)[0][0] = (vp_maxx - vp_minx) / 2.0;
   (*vp_matrix)[0][1] = 0.0;
   (*vp_matrix)[0][2] = Factor * ((*vp_matrix)[0][0] + vp_minx);
   (*vp_matrix)[0][3] = 0.0;

   (*vp_matrix)[1][0] = 0.0;
   (*vp_matrix)[1][1] = (vp_maxy - vp_miny) / 2.0;
   (*vp_matrix)[1][2] = Factor * ((*vp_matrix)[1][1] + vp_miny);
   (*vp_matrix)[1][3] = 0.0;

   (*vp_matrix)[2][0] = 0.0;
   (*vp_matrix)[2][1] = 0.0;
   (*vp_matrix)[2][2] = (((1.0 / (1.0 + Zmin)) + Factor) * (vp_maxz - vp_minz))
                     + (Factor * vp_minz);
   (*vp_matrix)[2][3] = vp_minz - vp_maxz;

   (*vp_matrix)[3][0] = 0.0;
   (*vp_matrix)[3][1] = 0.0;
   (*vp_matrix)[3][2] = Factor;
   (*vp_matrix)[3][3] = 0.0;

#ifdef DEBUG_VIEW
   printf("window to viewport matrix:\n");
   print_matrix(*vp_matrix);
#endif
}
